# coding=utf-8
# Copyright 2017 Flowdas Inc. <prospero@flowdas.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import absolute_import

import sys

PY2 = (sys.version_info[0] == 2)
PY3 = (sys.version_info[0] == 3)

import os
import logging
import signal
import importlib
import functools

import click
from pkg_resources import get_distribution

from flowdas.oliver import settings
from gevent import monkey
import yaml

__all__ = [
    '__author__',
    '__version__',
    'define',
    'command',
    'OptionType',
    'Path',
    'PY2',
    'PY3',
    'critical',
    'error',
    'warning',
    'info',
    'debug',
    'exception',
]

__author__ = u'오동권(Dong-gweon Oh) <prospero@flowdas.com>'
__version__ = getattr(get_distribution('flowdas-oliver'), 'version', None)

if PY3:
    def execfile(filename, globals):
        with open(filename) as f:
            source = f.read()
        code = compile(source, filename, 'exec')
        exec(code, globals)

_setup = False

_params = set()


class OliverfileNotFoundError(Exception):
    def __init__(self, oliverfile):
        super(OliverfileNotFoundError, self).__init__()
        self.oliverfile = oliverfile


def patch_falcon_cookies():
    import falcon

    def falcon_cookies(self):
        if self._cookies is None:
            cookietxt = self.get_header('Cookie')
            cookies = {}
            if cookietxt:
                for cook in cookietxt.split(';'):
                    cook = cook.lstrip()
                    try:
                        k, v = cook.split('=', 1)
                        if v.startswith('"') and v.endswith('"'):
                            v = v[1:-1]
                        cookies[k] = v
                    except ValueError:
                        pass
            self._cookies = cookies

        return self._cookies.copy()

    falcon.Request.cookies = property(falcon_cookies)


def setup(oliverfile, debug=False):
    global _setup
    if not _setup:
        _setup = True
        os.environ.setdefault('LC_CTYPE', 'ko_KR.UTF-8')
        monkey.patch_all(subprocess=True)
        patch_falcon_cookies()

        if oliverfile is None:
            for candidate in ['oliver.yaml', 'Oliverfile']:
                if os.path.isfile(candidate):
                    oliverfile = os.path.abspath(candidate)
                    break
            else:
                raise OliverfileNotFoundError('oliver.yaml')
        else:
            oliverfile = os.path.abspath(oliverfile)
            if not os.path.isfile(oliverfile):
                raise OliverfileNotFoundError(oliverfile)

        yaml_config = oliverfile.endswith('.yaml')

        if yaml_config:
            with open(oliverfile, 'rb') as f:
                context = yaml.load(f)
            if context is None:
                context = {}
        else:
            context = {
                'DEBUG': debug,
                'OLIVERFILE': oliverfile,
                '__file__': oliverfile,
            }
            execfile(oliverfile, context)
        for name in dir(settings):
            if name.startswith('_'):
                continue
            delattr(settings, name)
        for name in context:
            if name.startswith('_'):
                continue
            setattr(settings, name, context[name])
        define('DEBUG', debug, mutable=False)
        define('OLIVERFILE', oliverfile, mutable=False)
        define('PYTHONPATH', [], mutable=True, type=Path())
        define('ENV', {}, mutable=True)
        define('CHDIR', '.', mutable=True, type=Path())
        define('LOGLEVEL', 'debug' if debug else 'info', mutable=True)
        define('LOGFILE', 'Oliver.log', mutable=True, type=Path())
        if not yaml_config:
            define('main', lambda: None)
        sys.path.extend(settings.PYTHONPATH)
        oliver_path = os.path.dirname(settings.OLIVERFILE)
        retry_import = oliver_path not in sys.path
        os.environ.update(settings.ENV)
        if settings.CHDIR:
            os.chdir(settings.CHDIR)
        if yaml_config:
            if 'uses' in context:
                for modname in settings.uses:
                    try:
                        module = importlib.import_module(modname)
                    except ImportError:
                        if retry_import:
                            retry_import = False
                            sys.path.append(oliver_path)
                            module = importlib.import_module(modname)
                            settings.PYTHONPATH.append(oliver_path)
                        else:
                            raise
                    init_func = getattr(module, 'oliver_init', None)
                    if callable(init_func):
                        init_func()
        else:
            settings.main()


#
# setting parameters
#

class OptionType(object):
    def __call__(self, value):
        try:
            return self.convert(value)
        except:
            raise TypeError(self.fail(value))

    def convert(self, value):
        return value

    def fail(self, value):
        return str(value)


class Path(OptionType):
    def convert(self, value):
        def conv(val):
            return os.path.normpath(os.path.join(os.path.dirname(settings.OLIVERFILE), val))

        if isinstance(value, (list, tuple)):
            return tuple([conv(v) for v in value])
        else:
            return conv(value)


def define(name, default, mutable=True, type=None):
    """설정 파라미터를 정의한다.

    Since version 0.2.
    """
    if not _setup:
        raise ImportError('Oliver packages should be declared in a file named oliver.yaml.')
    _params.add(name)
    if not mutable or not hasattr(settings, name):
        setattr(settings, name, default)
    if type is not None:
        assert isinstance(type, OptionType)
        setattr(settings, name, type(getattr(settings, name)))


#
# logging
#

class Logger(object):
    LOG_LEVELS = {
        "critical": logging.CRITICAL,
        "error": logging.ERROR,
        "warning": logging.WARNING,
        "info": logging.INFO,
        "debug": logging.DEBUG
    }

    def __init__(self):
        self.log = logging.getLogger('oliver')
        self.log.propagate = False
        if settings.LOGFILE and settings.LOGFILE != '-':
            dirname = os.path.dirname(settings.LOGFILE)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            self.handler = logging.FileHandler(settings.LOGFILE)
            self.sighandler = signal.signal(signal.SIGUSR1, self.handle_usr1)
        else:
            self.handler = logging.StreamHandler(sys.stdout)
        logfmt = r"%(asctime)s [%(process)d] [%(levelname)s] %(message)s"
        datefmt = r"[%Y-%m-%d %H:%M:%S %z]"
        self.handler.setFormatter(logging.Formatter(logfmt, datefmt))
        self.log.addHandler(self.handler)
        self.log.setLevel(self.LOG_LEVELS.get(settings.LOGLEVEL, logging.INFO))

    def handle_usr1(self, sig, frame):
        if self.sighandler:
            self.sighandler(sig, frame)
        self.reopen_files()

    def reopen_files(self):
        handler = self.handler
        handler.acquire()
        try:
            if handler.stream:
                handler.stream.close()
                handler.stream = open(handler.baseFilename, handler.mode)
        finally:
            handler.release()


_logger = None


def logger():
    global _logger
    if _logger is None:
        _logger = Logger()
    return _logger


def critical(msg, *args, **kwargs):
    """*CRITICAL* 수준으로 메시지를 로그한다. 인자는 :py:func:`logging.critical` 처럼 해석된다.

    Since version 0.2.
    """
    logger().log.critical(msg, *args, **kwargs)


def error(msg, *args, **kwargs):
    """*ERROR* 수준으로 메시지를 로그한다. 인자는 :py:func:`logging.error` 처럼 해석된다.

    Since version 0.2.
    """
    logger().log.error(msg, *args, **kwargs)


def warning(msg, *args, **kwargs):
    """*WARNING* 수준으로 메시지를 로그한다. 인자는 :py:func:`logging.warning` 처럼 해석된다.

    Since version 0.2.
    """
    logger().log.warning(msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    """*INFO* 수준으로 메시지를 로그한다. 인자는 :py:func:`logging.info` 처럼 해석된다.

    Since version 0.2.
    """
    logger().log.info(msg, *args, **kwargs)


def debug(msg, *args, **kwargs):
    """*DEBUG* 수준으로 메시지를 로그한다. 인자는 :py:func:`logging.debug` 처럼 해석된다.

    Since version 0.2.
    """
    logger().log.debug(msg, *args, **kwargs)


def exception(msg, *args, **kwargs):
    """*ERROR* 수준으로 메시지를 로그한다. 인자는 :py:func:`logging.exception` 처럼 해석된다.

    Since version 0.2.
    """
    logger().log.exception(msg, *args, **kwargs)


#
# command line interface
#

class Cli(click.Group):
    def invoke(self, ctx):
        try:
            setup(ctx.params['file'], ctx.params['debug'])
        except OliverfileNotFoundError as e:
            click.echo("No Oliverfile found: %s\n" % e.oliverfile)
            click.echo(ctx.get_help(), color=ctx.color)
            ctx.exit()
        super(Cli, self).invoke(ctx)


@click.command(cls=Cli, invoke_without_command=True)
@click.option('-f', '--file', type=click.Path('rb'), metavar='FILE', help='Read FILE as an Oliverfile.')
@click.option('-d', '--debug', is_flag=True, help='Enable debugging outputs.')
@click.pass_context
def cli(ctx, file, debug):
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help(), color=ctx.color)
        ctx.exit()


def command(*args, **kwargs):
    """명령행 인터페이스에 명령을 추가하는 데코레이터.

    인자는 :py:func:`click.command` 의 정의를 따른다.

    Example:

        >>> from flowdas import oliver
        >>> import click
        >>>
        >>> @oliver.command(help='Echo text')
        ... @click.argument('text')
        ... def echo(text):
        ...     click.echo(text)

    Since version 0.2.
    """

    def deco(func):
        return cli.command(*args, **kwargs)(func)

    return deco


@command(help='Print version string.')
def version():
    click.echo(__version__)


@command(name='settings', help='Show settings.')
def _settings():
    for name in sorted(dir(settings)):
        if name in _params:
            value = getattr(settings, name)
            if callable(value):
                modname = value.__module__
                if not modname:
                    modname = 'OLIVERFILE'
                click.echo('%s = %s.%s' % (name, modname, value.__name__))
            else:
                click.echo('%s = %r' % (name, value))


def _timeit(fut, number, repeat):
    import timeit
    context = {'fut': fut}
    best = min(timeit.repeat('fut()', repeat=repeat, number=number, globals=context))
    usec = best * 1e6 / number
    msgs = []
    msgs.append("%d loops," % number)
    precision = 3
    units = {"usec": 1, "msec": 1e3, "sec": 1e6}
    scales = [(scale, unit) for unit, scale in units.items()]
    scales.sort(reverse=True)
    for scale, time_unit in scales:
        if usec >= scale:
            break
    msgs.append("best of %d: %.*g %s per loop" % (repeat, precision, usec / scale, time_unit))
    click.echo(' '.join(msgs))


@command(help='Time execution of a command.')
@click.option('-n', '--number', default=1, help='Execute the given command INTEGER times in a loop.')
@click.option('-r', '--repeat', default=3, help='Repeat the loop iteration INTEGER times and take the best result. [3]')
@click.argument('command')
@click.argument('args', nargs=-1)
def timeit(number, repeat, command, args):
    from .testing import simulate_command
    fut = functools.partial(simulate_command, *((command,) + args))
    _timeit(fut, number=number, repeat=repeat)


def _profile(fut, output, number):
    import cProfile
    def run():
        for i in range(number):
            fut()

    context = {'run': run}
    cProfile.runctx('run()', globals=context, locals=context, filename=output)


@command(help='Time execution of a command.')
@click.option('-o', '--output', default='Oliver.prof', help='Output file.')
@click.option('-n', '--number', default=1, help='Execute the given command INTEGER times in a loop.')
@click.argument('command')
@click.argument('args', nargs=-1)
def profile(output, number, command, args):
    from click.testing import CliRunner

    runner = CliRunner()
    fut = functools.partial(runner.invoke, cli, (command,) + args)
    _profile(fut, output=output, number=number)


def main():
    cli()
